
Function ImportEx-VM
{# .ExternalHelp  MAML-VM.XML
    [CmdletBinding(SupportsShouldProcess=$True)]
    Param (
    
        [parameter(Mandatory = $true, ValueFromPipelineByPropertyName =$true,ValueFromPipeline =$true)][Alias("Fullname","Path","DiskPath")][ValidateNotNullOrEmpty()]
        [string[]]$Paths,

        [string]$CloneRoot,

        [string]$Name,
        
        [ValidateNotNullOrEmpty()]
        [String]$Server=".",   #Only import on one server 
        
        [switch]$ReimportVM, 
        [switch]$ReuseIDs,  
        [switch]$Wait, 
        [Switch]$Preserve,
        
        $PSC, 
        [switch]$Force
    )
    process {
        if ($psc -eq $null)  {$psc = $pscmdlet} ; if (-not $PSBoundParameters.psc) {$PSBoundParameters.add("psc",$psc)}
        forEach ($Path in $Paths) {
            if ($server -eq ".") {$path = (Resolve-Path $path).path}

            if ($pscmdlet.shouldProcess(($lStr_ImportFrom -f $server))) {
                if ($Preserve) {Add-ZIPContent "$path\importFiles.zip" "$Path\config.xml","$path\virtual machines"
                                if (test-path "$path\$($vm.elementname)\snapshots") {Add-ZIPContent "$path\importFiles.zip" "$path\snapshots"}
                }  
                elseif ($reImportVM -and (Test-Path "$path\importFiles.zip")) {
                    Copy-ZipContent -Zipfile "$path\importFiles.zip" -path $path
                    Remove-VM -VM $ReimportVM -Server $server -confirm
                }

                $VSMgtSvc = Get-WmiObject -ComputerName $server -Namespace $HyperVNamespace -Class "MsVM_virtualSystemManagementService"

                $inParams = $VSMgtSvc.GetVirtualSystemImportSettingData($path)
                $inParams = $inParams.ImportSettingData
                $inParams.GenerateNewId = -not $ReuseIDs.Ispresent;
                $inParams.CreateCopy = -not $ReuseIDs.Ispresent;
                $inParams.Name = $Name;

                $importDir = Join-Path $CloneRoot $Name
                $inParams.TargetVhdDataRoot = $importDir
                $inParams.TargetSnapshotDataRoot = $importDir
                $inParams.TargetVmDataRoot = $importDir
                if (!(Test-Path -path "$importDir")) {
                    New-Item "$importDir" -type directory
                }

                $Result = $VSMgtSvc.importVirtualSystemEx($Path, $inParams.GetText(1))
                If ($wait) {
                    $job = test-wmijob $Result.Job -wait -Description ($lstr_importFrom -f $path)
                    if ($job.JobState -eq [JobState]::Completed) { 
                        Write-Verbose $lstr_ImportSuccess 
                        return $TRUE
                    } else { 
                        Write-Error $Job.errorDescription 
                        return $FALSE
                    }
                }           
                else {
                    Write-Verbose ([ReturnCode]$result.returnValue) 
                    $Result.Job
                }
            }
        }
    }     
}
