<###################################################################################
#                                                                                  #
#  Parallels Plesk Panel Windows Server 2008 R2 Hyper-V VM cloning script example. #
#  WARNING: This is example only, please check this script before using.           #
#                                                                                  #
####################################################################################>

Import-Module .\HyperV\disk.ps1
Import-Module .\HyperV\Helper.ps1
Import-Module .\HyperV\menu.ps1
Import-Module .\HyperV\Network.ps1
Import-Module .\HyperV\v1Compatability.ps1
Import-Module .\HyperV\vm.ps1
Import-Module .\HyperV\VMConfig.ps1
Import-Module .\HyperV\En-us.ps1

Import-Module .\CloneVM-Utils.ps1

if ( $args.Count -lt 5 )
{
    Write-Host
    Write-host("Usage: .\CloneVM.ps1 <Importing image path> <Root path for cloned VMs> <New VM Hyper-V name> <New VM Host name> <New VM Administrator password> [New VM MAC (optinal)]")
    Write-Host
    Write-host("Example: .\CloneVM.ps1 `"C:\PPP 11.0 Image`" `"C:\PPP 11.0 Clones`" `"PLESK VM 001`" `"HVPLESK001`" `"1qazxsW`" `"00155DECF690`"")
    Write-Host
    Read-host("Press [Enter] to start 'PowerShell management Library for Hyper-V' interactive menu")

    Show-HypervMenu
    return $FALSE
}

# Parameter values:
$VMImagePath = $args[0]
$VMCloneRootPath = $args[1]
$VMHyperVName = $args[2]
$VMHostName = $args[3]
$VMAdministratorPassword = $args[4]
$VMMAC = $args[5]

# Cloning defaults:
$VMDefaultLogin = "Administrator"   # VM image administrator login
$VMDefaultPassword = "1q2w3e4R"     # VM image administrator password
$VMServer = "."                     # change this value to clone VM on another Hyper-V server
$VMHeartBeatTimeOut = 600           # maximum timeout while Hyper-v detects that VM is running
$VMStaringTimeOut = 20              # timeout for Windows services start after "heart beat" is detected
$VMPingAttempts = 100               # number of attempts for DHCP IP resolving
$VMRPCAttempts = 60                 # number of attempts to rerun VM preparing script with RPC error
$VMStartAttempts = 60               # number of attempts to restart VM after shutdown

Write-Host
Write-Host("=================================================")
Write-Host("Begin Hyper-V VM cloning with parameters:")
Write-Host(" - Importing image path: '$VMImagePath'")
Write-Host(" - Root path for cloned VMs: '$VMCloneRootPath'")
Write-Host(" - New VM Hyper-V name: '$VMHyperVName'")
Write-Host(" - New VM Host name: '$VMHostName'")
Write-Host(" - New VM Administrator password: '$VMAdministratorPassword'")
Write-Host(" - New VM MAC (optinal): '$VMMAC'")
Write-Host("=================================================")
date
Write-Host

Write-Host("Importing new VM from '$VMImagePath' to '$VMCloneRootPath',")
Write-Host("New VM Hyper-V name: '$VMHyperVName'...")
$Result = ImportEx-VM -Path "$VMImagePath" -CloneRoot "$VMCloneRootPath" -Name "$VMHyperVName" -Server "$VMServer" -Wait -ReuseIDs:$false
if ( ! $Result ) 
{
    Write-Host("Exit with error.")
    return $FALSE
}
Write-Host

# Recreate NIC to avoid "Get-WmiObject : Generic failure" error
# at getting IP address of cloned VM.

Write-Host("Recreating first VM NIC.")
Write-Host("Determinating first VM NIC...")
$VMNIC = Get-VMNIC -Server "$VMServer" -VM "$VMHyperVName" | Select-Object -First 1
if ( $VMNIC ) 
{
    Write-Host("Removing VM NIC '{0}'..." -f (Get-VMNICSwitch -NIC $VMNIC).ElementName)
    $Result = Remove-VMNIC -Server "$VMServer" -VM "$VMHyperVName" -NIC $VMNIC
    if ( $Result -ne [ReturnCode]::OK )
    {
        Write-Host("Exit with error.")
        return $FALSE
    }
} 
else 
{
    Write-Host("VM NIC is not found.")
}

Write-Host("Creating new VM NIC...")
$switch = Select-VMSwitch -Server "$VMServer"
$VMNIC = Add-VMNIC -Server "$VMServer" -VM "$VMHyperVName" -legacy:$FALSE -VirtualSwitch $switch
if ( ! $VMNIC )
{
    Write-Host("Exit with error.")
    return $FALSE
}
Write-Host

if ( $VMMAC )
{
    Write-Host("Changing VM MAC to: '$VMMAC'...")
    $Result = Set-VMNICAddress -Server "$VMServer" -VM "$VMHyperVName" -NIC $VMNIC -MAC "$VMMAC" -Force
    if ( ! $Result )
    {
        Write-Host("Exit with error.")
        return $FALSE
    }
    Write-Host
}

Write-Host("Starting VM...")
$Result = Start-VM -Server "$VMServer" -VM "$VMHyperVName" -Wait -HeartBeatTimeOut $VMHeartBeatTimeOut
if ( ($Result[0] -ne [ReturnCode]::OK) -or ($Result[1].Status -ne [HeartBeatICStatus]::OK) )
{
    $Result
    Write-Host("Exit with error.")
    return $FALSE
}
Start-Sleep -Seconds $VMStaringTimeOut
Write-Host

Write-Host("Getting VM IP address...")
for ($i=0; $i -lt $VMPingAttempts; $i++) 
{ 
    $VMIP = (Ping-VM -Server "$VMServer" -VM "$VMHyperVName" -UseIP4Address).NetworkAddress    
    if ( $VMIP ) { 
        break 
    }

    Write-Host "." -NoNewLine
    Start-Sleep -Seconds 1
}
if ( ! $VMIP ) 
{
    Write-Host
    Write-Host("Exit with error.")
    return $FALSE
}
Write-Host
Write-Host("New VM IP address: '$VMIP'")
Write-Host

# PsExec arguments
$VMPreparingArgumentList = 
    "\\$VMIP -h -n 180 -u `"$VMDefaultLogin`" -p `"$VMDefaultPassword`" -c -f .\prepare_vm.cmd `"$VMAdministratorPassword`" `"$VMHostName`""

Write-Host("Initializing Parallels Plesk Panel environment, enabling Remote Desktop `(RDP`),")
Write-Host("Changing Administrator password, Host name and install Parallels Plesk Panel Micro Updates.")
Write-Host("PsExec arguments: '$VMPreparingArgumentList'.")
for ($i=0;; $i++) 
{ 
    Write-Host "Prepare VM executing try: $($i+1)... "
    $Result = Start-Process -NoNewWindow -Wait -PassThru -FilePath ".\psexec.exe" -ArgumentList $VMPreparingArgumentList
    Write-Host
    if ( [int]$Result.ExitCode -eq 0 ) {
        break
    }

    # If "The RPC server is unavailable" error
    if ( [int]$Result.ExitCode -eq 1722 ) 
    { 
        Write-Host("`"The RPC server is unavailable`" error detected.")

        if ( $i -lt $VMRPCAttempts ) 
        {
            Start-Sleep -Seconds 10
            continue
        }
    }

    Write-Host("VM preparing script is not executed properly. Exit code: {0}." -f [int]$Result.ExitCode)
    Write-Host("Shutdowning VM...")
    Invoke-VMShutdown -Server "$VMServer" -VM "$VMHyperVName" -ShutdownTimeOut $VMHeartBeatTimeOut -Force
    Write-Host("Exit with error.")
    return $FALSE
}

Write-Host("Restarting VM. Shutdowning...")
Invoke-VMShutdown -Server "$VMServer" -VM "$VMHyperVName" -ShutdownTimeOut $VMHeartBeatTimeOut -Force
for ($i=0;; $i++) 
{ 
    Write-Host "Starting VM try: $($i+1)... "
    $Result = Start-VM -Server "$VMServer" -VM "$VMHyperVName" -Wait -HeartBeatTimeOut $VMHeartBeatTimeOut
    if ( $Result[0] -ne [ReturnCode]::OK ) 
    {
        if ( $i -eq $VMStartAttempts ) 
        {
            Write-Host
            Write-Host("Exit with error.")
            return $FALSE
        }

        Start-Sleep -Seconds 10
        continue
    }

    break
}
Write-Host

Write-Host("New VM summary:")
Get-VMSummary -Server "$VMServer" -VM "$VMHyperVName"
Write-Host

date
Write-Host("Exit with success.")
return $TRUE
